@@ -39,12 +39,23 @@
 #include <sys/mman.h>
 #include <sys/syscall.h>  /* for SYS_mmap */
 
-#include <algorithm>
-#include <string>
-#include <set>
-#include <vector>
 #include <string.h>
 
+// XXX: it seems setting macro in CMakeLists.txt does not work,
+// so manually set it here now.
+
+// Building msandr client for running in DynamoRIO hybrid mode,
+// which allows some module running natively.
+// TODO: turn it on by default when hybrid is stable enough
+// #define MSANDR_NATIVE_EXEC
+
+#ifndef MSANDR_NATIVE_EXEC
+#include <algorithm>
+#include <set>
+#include <string>
+#include <vector>
+#endif
+
 #define TESTALL(mask, var) (((mask) & (var)) == (mask))
 #define TESTANY(mask, var) (((mask) & (var)) != 0)
 
@@ -60,14 +71,6 @@
 
 #define VERBOSITY 0
 
-// XXX: it seems setting macro in CMakeLists.txt does not work,
-// so manually set it here now.
-
-// Building msandr client for running in DynamoRIO hybrid mode,
-// which allows some module running natively.
-// TODO: turn it on by default when hybrid is stable enough
-// #define MSANDR_NATIVE_EXEC
-
 // Building msandr client for standalone test that does not need to
 // run with msan build executables. Disable by default.
 // #define MSANDR_STANDALONE_TEST
@@ -89,9 +92,11 @@
 # define SHADOW_MEMORY_MASK 0x3fffffffffffULL
 #endif /* MSANDR_STANDALONE_TEST */
 
-namespace {
+typedef void *(*WrapperFn)(void *);
+extern "C" void __msan_set_indirect_call_wrapper(WrapperFn wrapper);
+extern "C" void __msan_dr_is_initialized();
 
-std::string g_app_path;
+namespace {
 
 int msan_retval_tls_offset;
 int msan_param_tls_offset;
@@ -186,7 +191,6 @@ void InitializeMSanCallbacks() {
               dr_get_application_name());
     CHECK(app);
   }
-  g_app_path = app->full_path;
 
   __msan_get_retval_tls_offset = (int (*)())
       LookupCallback(app, "__msan_get_retval_tls_offset");
@@ -366,8 +370,18 @@ void InstrumentReturn(void *drcontext, instrlist_t *bb, instr_t *instr) {
                                        OPSZ_PTR),
              OPND_CREATE_INT32(0)));
 #else  /* !MSANDR_STANDALONE_TEST */
+# ifdef MSANDR_NATIVE_EXEC
+  /* For optimized native exec, -mangle_app_seg and -private_loader are turned off,
+   * so we can reference msan_retval_tls_offset directly.
+   */
+  PRE(instr,
+      mov_st(drcontext,
+             opnd_create_far_base_disp(DR_SEG_FS, DR_REG_NULL, DR_REG_NULL, 0,
+                                       msan_retval_tls_offset, OPSZ_PTR),
+             OPND_CREATE_INT32(0)));
+# else /* !MSANDR_NATIVE_EXEC */
   /* XXX: the code below only works if -mangle_app_seg and -private_loader, 
-   * which is turned of for optimized native exec
+   * which is turned off for optimized native exec
    */
   dr_save_reg(drcontext, bb, instr, DR_REG_XAX, SPILL_SLOT_1);
 
@@ -382,7 +396,7 @@ void InstrumentReturn(void *drcontext, instrlist_t *bb, instr_t *instr) {
              OPND_CREATE_INT32(0)));
 
   dr_restore_reg(drcontext, bb, instr, DR_REG_XAX, SPILL_SLOT_1);
-
+# endif /* !MSANDR_NATIVE_EXEC */
   // The original instruction is left untouched. The above instrumentation is just
   // a prefix.
 #endif  /* !MSANDR_STANDALONE_TEST */
@@ -403,6 +417,16 @@ void InstrumentIndirectBranch(void *drcontext, instrlist_t *bb,
                  OPND_CREATE_INT32(0)));
   }
 #else  /* !MSANDR_STANDALONE_TEST */
+# ifdef MSANDR_NATIVE_EXEC
+  for (int i = 0; i < NUM_TLS_PARAM; ++i) {
+    PRE(instr,
+        mov_st(drcontext,
+               opnd_create_far_base_disp(DR_SEG_FS, DR_REG_NULL, DR_REG_NULL, 0,
+                                         msan_param_tls_offset + i*sizeof(void*),
+                                         OPSZ_PTR),
+               OPND_CREATE_INT32(0)));
+  }
+# else /* !MSANDR_NATIVE_EXEC */
   /* XXX: the code below only works if -mangle_app_seg and -private_loader, 
    * which is turned off for optimized native exec
    */
@@ -422,7 +446,7 @@ void InstrumentIndirectBranch(void *drcontext, instrlist_t *bb,
   }
 
   dr_restore_reg(drcontext, bb, instr, DR_REG_XAX, SPILL_SLOT_1);
-
+# endif /* !MSANDR_NATIVE_EXEC */
   // The original instruction is left untouched. The above instrumentation is just
   // a prefix.
 #endif  /* !MSANDR_STANDALONE_TEST */
@@ -792,6 +816,8 @@ DR_EXPORT void dr_init(client_id_t id) {
   drmgr_init();
   drutil_init();
 
+#ifndef MSANDR_NATIVE_EXEC
+  // We should use drconfig to ignore these applications.
   std::string app_name = dr_get_application_name();
   // This blacklist will still run these apps through DR's code cache.  On the
   // other hand, we are able to follow children of these apps.
@@ -802,6 +828,7 @@ DR_EXPORT void dr_init(client_id_t id) {
       app_name == "sh" || app_name == "true" || app_name == "exit" ||
       app_name == "yes" || app_name == "echo")
     return;
+#endif /* !MSANDR_NATIVE_EXEC */
 
   drsys_options_t ops;
   memset(&ops, 0, sizeof(ops));
@@ -866,6 +893,8 @@ DR_EXPORT void dr_init(client_id_t id) {
   drmgr_register_module_load_event(event_module_load);
   drmgr_register_module_unload_event(event_module_unload);
 #endif /* MSANDR_NATIVE_EXEC */
+  __msan_dr_is_initialized();
+  __msan_set_indirect_call_wrapper(dr_app_handle_mbr_target);
   if (VERBOSITY > 0)
     dr_printf("==MSANDR== Starting!\n");
 }
